#    Oak - a single-system message-board
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
Oak dbm backend

Implements the Oak message-board storage with dbm files
"""

import os
import anydbm
import time

class Tree:
    """The tree represents the overall sytem of oak boards"""

    def __init__(self, root=None, spliton=None, ext='.db'):
        if root is None:
            raise AttributeError("oak/dbm/__init__.py: root may not be none")
        if not os.path.exists(root):
            raise AttributeError("oak/dbm/__init__.py: root doesn't exist")
        self._root = root
        self._split = spliton
        self._ext = ext

    def trunk_file(self, trunk):
        """Return filename for a given trunkname, root, and split config"""

        if os.path.splitdrive(trunk)[0] or trunk.startswith(os.sep):
            return os.path.normpath(trunk)
        elif self._split:
            dirs = [self._root]
            dirs.extend(trunk.split(self._split))
            file = os.path.normpath(os.path.join(*dirs))
        else:
            file = os.path.normpath(os.path.join(self._root, trunk))
        # print 'trunk_file:', trunk, '->', file + self._ext
        return file + self._ext

    def new_trunk(self, trunk):
        """Create a new trunk.  Requires admin access."""

        file = self.trunk_file(trunk)
        if not os.path.exists(os.path.dirname(file)):
            os.makedirs(os.path.dirname(file))
        print 'OAK: Creating trunk', trunk, 'in file', file
        db = anydbm.open(file, 'c')
        if db.has_key('created'):
            raise ValueError("Trunk '%s' (file=%s) already exists." %
                (trunk,file))
        db['created'] = str(time.time())
        self.reset_trunk(db=db)
        db.close()

    def reset_trunk(self, trunk=None, db=None):
        """reset_trunk(self, trunk=None, db=None) -> None
        
        Reset a trunk to new empty status.  Irreversible.  db takes precendence
        over trunk."""

        close = 0
        if db is None:
            close = 1
            if trunk is None:
                raise ValueError("Neither trunk nor db specified")
            file = self.trunk_file(trunk)
            db = anydbm.open(file, 'w')

        for key in db.keys():
            if key != 'created':
                del db[key]

        db['reset'] = str(time.time())
        db['branches'] = str(0)
        db['modified'] = str(time.time())
        
        # TODO: set default tree properties here
        # (for now, nothing else)

        if close: db.close()

    def trunk(self, trunk):
        """trunk(self, trunk) -> Trunk

        Return a trunk object for a named trunk."""

        file = self.trunk_file(trunk)
        try:
            return Trunk(file)
        except anydbm.error:
            raise ValueError("Trunk %s (%s) could not be opened" % (trunk, file))

class Trunk:
    """A trunk is a named board"""

    def __init__(self, file):
        self._file = file
        try:
            self._db = anydbm.open(file, 'r')
        except anydbm.error, e:
            self._db = None
            raise e
        
        # TODO: verify permissions - the most specific one wins
        # (for now, assume full perms)
        #
        # This will look like
        #   if db.has_key('perms_user_%s' % user.id):
        #       ...
        #   else:
        #       for group in user.groups():
        #           if db.has_key('perms_group_%s' % group.id):
        #               ...
        #               break
        #       else:
        #           assume no perms

    def __del__(self):
        if self._db is not None: self._db.close()

    def branch_count(self):
        return self._db['branches']

    def branch(self, num=None):
        if num is None:
            return Branch(self)
        elif 0 <= int(num) < int(self._db['branches']):
            return Branch(self, int(num))
        else:
            raise IndexError("Branch index %s out of range" % num)

    def date(self, type):
        if type in ('modified', 'created', 'reset'):
            return float(self._db[type])
        else:
            raise KeyError("There is no date '%s'" % type)

    def writedb(self):
        dbw = anydbm.open(self._file, 'w')
        dbw['modified'] = str(time.time())
        return dbw

class Branch:
    """A branch is a string of messages"""

    def __init__(self, trunk, num=None):
        self._trunk = trunk # don't close db prematurely, etc.
        self._db = trunk._db
        self._num = num
        self._title = None

    def leaf_count(self):
        return self._db['branch_%s_leaves' % self._num]

    def leaf(self, num):
        if 0 <= int(num) < int(self._db['branch_%s_leaves' % self._num]):
            return Leaf(self, int(num))
        else:
            raise IndexError("Leaf index %s out of range" % num)

    def new(self, title, fromstring=None, fromfile=None):

        leaf = Leaf(self, 0)
        leaf.message(fromstring, fromfile)

        dbw = self._trunk.writedb()
        dbw2 = self._trunk.writedb()
        self._num = branch = int(dbw['branches'])
        self._title = title

        # Note: don't store dbw in self; that'd keep exclusive write open
        dbw['branches'] = str(branch+1)
        dbw['branch_%s_title' % branch] = title
        dbw['branch_%s_leaves' % branch] = str(0)
        leaf.append(dbw)
        del dbw

    def title(self, title=None):
        if title is not None:
            self._title = title
            self._trunk.writedb()['branch_%s_title' % self._num] = title
        elif self._title is not None:
            return self._title
        else:
            return self._db['branch_%s_title' % self._num]

class Leaf:
    """A leaf is a message"""

    def __init__(self, branch, num=None):
        self._branch = branch
        self._db = branch._db
        self._num = num
        self._meta = None
        self._author = 'Anonymous'
        self._auid = -1
        self._date = time.time()

    def append(self, dbw):
        branch = self._branch._num
        self._num = leaves = int(dbw['branch_%s_leaves' % branch])
        dbw['branch_%s_leaves' % branch] = str(leaves+1)
        dbw['branch_%s_leaf_%s_meta' % (branch, leaves)] = self.meta()
        dbw['branch_%s_leaf_%s_data' % (branch, leaves)] = self.message()

    def read(self):
        if self._num is None:
            raise ValueError("Cannot read an unposted leaf")
        try:
            meta = self._db['branch_%s_leaf_%s_meta' % (self._branch._num, self._num)]
            data = self._db['branch_%s_leaf_%s_data' % (self._branch._num, self._num)]
        except KeyError:
            raise IndexError("Leaf removed")

        self._meta = meta
        author, auid, date = meta.split('::')
        self._author = author
        self._auid = int(auid)
        self._date = float(date)
        self._message = data

    def author(self, newauthor=None):
        if newauthor:
            self._author = newauthor
            self._meta = None
            return None
        else:
            return self._author

    def auid(self, newauid=None):
        if newauid:
            self._auid = newauid
            self._meta = None
            return None
        else:
            return self._auid

    def date(self, newdate=None):
        if newdate:
            self._date = newdate
            self._meta = None
            return None
        else:
            return self._date

    def meta(self):
        if not self._meta:
            self._meta = '::'.join(map(str, (self._author, self._auid, self._date)))
        return self._meta

    def message(self, fromstring=None, fromfile=None):
        if fromstring is not None:
            self._message = fromstring
            return None
        elif fromfile is not None:
            file = open(fromfile, r)
            self._message = '\n'.join(file.readlines())
            return None
        else:
            return self._message
